@@ -20,8 +20,11 @@
 
 """Miscellaneous utility classes and functions"""
 
+import asyncio
+import fnmatch
 import functools
 import ipaddress
+import os
 import re
 import shlex
 import socket
@@ -31,8 +34,8 @@ from pathlib import Path, PurePath
 from random import SystemRandom
 from types import TracebackType
 from typing import Any, AsyncContextManager, Awaitable, Callable, Dict
-from typing import Generator, Generic, IO, Mapping, Optional, Sequence
-from typing import Tuple, Type, TypeVar, Union, cast, overload
+from typing import Generator, Generic, IO, Iterator, Mapping, Optional
+from typing import Sequence, Tuple, Type, TypeVar, Union, cast, overload
 from typing_extensions import Literal, Protocol
 
 from .constants import DEFAULT_LANG
@@ -108,6 +111,10 @@ IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
 IPNetwork = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
 SockAddr = Union[Tuple[str, int], Tuple[str, int, int, int]]
 
+EnvMap = Mapping[BytesOrStr, BytesOrStr]
+EnvItems = Sequence[Tuple[BytesOrStr, BytesOrStr]]
+EnvSeq = Sequence[BytesOrStr]
+Env = Optional[Union[EnvMap, EnvItems, EnvSeq]]
 
 # Define a version of randrange which is based on SystemRandom(), so that
 # we get back numbers suitable for cryptographic use.
@@ -120,6 +127,62 @@ _time_units = {'': 1, 's': 1, 'm': 60, 'h': 60*60,
                'd': 24*60*60, 'w': 7*24*60*60}
 
 
+def encode_env(env: Env) -> Iterator[Tuple[bytes, bytes]]:
+    """Convert environemnt dict or list to bytes-based dictionary"""
+
+    env = cast(Sequence[Tuple[BytesOrStr, BytesOrStr]],
+               env.items() if isinstance(env, dict) else env)
+
+    try:
+        for item in env:
+            if isinstance(item, (bytes, str)):
+                if isinstance(item, str):
+                    item = item.encode('utf-8')
+
+                key_bytes, value_bytes = item.split(b'=', 1)
+            else:
+                key, value = item
+
+                key_bytes = key.encode('utf-8') \
+                    if isinstance(key, str) else key
+
+                value_bytes = value.encode('utf-8') \
+                    if isinstance(value, str) else value
+
+            yield key_bytes, value_bytes
+    except (TypeError, ValueError) as exc:
+        raise ValueError(f'Invalid environment value: {exc}') from None
+
+
+def lookup_env(patterns: EnvSeq) -> Iterator[Tuple[bytes, bytes]]:
+    """Look up environemnt variables with wildcard matches"""
+
+    for pattern in patterns:
+        if isinstance(pattern, str):
+            pattern = pattern.encode('utf-8')
+
+        if os.supports_bytes_environ:
+            for key_bytes, value_bytes in os.environb.items():
+                if fnmatch.fnmatch(key_bytes, pattern):
+                    yield key_bytes, value_bytes
+        else: # pragma: no cover
+            for key, value in os.environ.items():
+                key_bytes = key.encode('utf-8')
+                value_bytes = value.encode('utf-8')
+                if fnmatch.fnmatch(key_bytes, pattern):
+                    yield key_bytes, value_bytes
+
+
+def decode_env(env: Dict[bytes, bytes]) -> Iterator[Tuple[str, str]]:
+    """Convert bytes-based environemnt dict to Unicode strings"""
+
+    for key, value in env.items():
+        try:
+            yield key.decode('utf-8'), value.decode('utf-8')
+        except UnicodeDecodeError:
+            pass
+
+
 def hide_empty(value: object, prefix: str = ', ') -> str:
     """Return a string with optional prefix if value is non-empty"""
 
@@ -130,7 +193,7 @@ def hide_empty(value: object, prefix: str = ', ') -> str:
 def plural(length: int, label: str, suffix: str = 's') -> str:
     """Return a label with an optional plural suffix"""
 
-    return '%d %s%s' % (length, label, suffix if length != 1 else '')
+    return f'{length} {label}{suffix if length != 1 else ""}'
 
 
 def all_ints(seq: Sequence[object]) -> bool:
@@ -356,6 +419,14 @@ async def maybe_wait_closed(writer: '_SupportsWaitClosed') -> None:
         pass
 
 
+async def run_in_executor(func: Callable[..., _T], *args: object) -> _T:
+    """Run a function in an asyncio executor"""
+
+    loop = asyncio.get_event_loop()
+
+    return await loop.run_in_executor(None, func, *args)
+
+
 def set_terminal_size(tty: IO, width: int, height: int,
                       pixwidth: int, pixheight: int) -> None:
     """Set the terminal size of a TTY"""
@@ -372,8 +443,8 @@ class Options:
     def __init__(self, options: Optional['Options'] = None, **kwargs: object):
         if options:
             if not isinstance(options, type(self)):
-                raise TypeError('Invalid %s, got %s' %
-                                (type(self).__name__, type(options).__name__))
+                raise TypeError(f'Invalid {type(self).__name__}, '
+                                f'got {type(options).__name__}')
 
             self.kwargs = options.kwargs.copy()
         else:
@@ -385,7 +456,7 @@ class Options:
     def prepare(self, **kwargs: object) -> None:
         """Pre-process configuration options"""
 
-    def update(self, kwargs: Dict[str, object]) -> None:
+    def update(self, **kwargs: object) -> None:
         """Update options based on keyword parameters passed in"""
 
         self.kwargs.update(kwargs)
@@ -424,15 +495,15 @@ class Record(metaclass=_RecordMeta):
             setattr(self, k, v)
 
     def __repr__(self) -> str:
-        return '%s(%s)' % (type(self).__name__,
-                           ', '.join('%s=%r' % (k, getattr(self, k))
-                                     for k in self.__slots__))
+        values = ', '.join(f'{k}={getattr(self, k)!r}' for k in self.__slots__)
+
+        return f'{type(self).__name__}({values})'
 
     def __str__(self) -> str:
         values = ((k, self._format(k, getattr(self, k)))
                   for k in self.__slots__)
 
-        return ', '.join('%s: %s' % (k, v) for k, v in values if v is not None)
+        return ', '.join(f'{k}: {v}' for k, v in values if v is not None)
 
     def _format(self, k: str, v: object) -> Optional[str]:
         """Format a field as a string"""
@@ -717,7 +788,7 @@ class PasswordChangeRequired(Exception):
     """
 
     def __init__(self, prompt: str, lang: str = DEFAULT_LANG):
-        super().__init__('Password change required: %s' % prompt)
+        super().__init__(f'Password change required: {prompt}')
         self.prompt = prompt
         self.lang = lang
 
@@ -735,7 +806,7 @@ class BreakReceived(Exception):
     """
 
     def __init__(self, msec: int):
-        super().__init__('Break for %s msec' % msec)
+        super().__init__(f'Break for {msec} msec')
         self.msec = msec
 
 
@@ -752,7 +823,7 @@ class SignalReceived(Exception):
     """
 
     def __init__(self, signal: str):
-        super().__init__('Signal: %s' % signal)
+        super().__init__(f'Signal: {signal}')
         self.signal = signal
 
 
@@ -790,8 +861,8 @@ class TerminalSizeChanged(Exception):
     """
 
     def __init__(self, width: int, height: int, pixwidth: int, pixheight: int):
-        super().__init__('Terminal size change: (%s, %s, %s, %s)' %
-                         (width, height, pixwidth, pixheight))
+        super().__init__(f'Terminal size change: ({width}, {height}, '
+                         f'{pixwidth}, {pixheight})')
         self.width = width
         self.height = height
         self.pixwidth = pixwidth
@@ -824,4 +895,4 @@ def construct_disc_error(code: int, reason: str, lang: str) -> DisconnectError:
     try:
         return _disc_error_map[code](reason, lang)
     except KeyError:
-        return DisconnectError(code, '%s (error %d)' % (reason, code), lang)
+        return DisconnectError(code, f'{reason} (error {code})', lang)
